unit sHintManager;
(************************************************************************
sStyle components package
Copyright (c) 2002 Serge Goncharov
                            Ukraine, Odessa
                            http://www.geocities.com/sstylish2002/
                            Serge@HyBce.com
 Version:	1.1, 04-02-2002
*************************************************************************)

{$DEFINE ALLOWPROP}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, sStyleUtil,
  sGraphUtils, sConst, sVclUtils, sUtils, ExtCtrls, sMessages;

type
  TsHintStyle = (hsSimply, {hsExtended,} hsStandard, hsNone);
  TsMousePosition = (mpLeftTop, mpLeftBottom, mpRightTop, mpRightBottom);

  TsPassivePaintStyle = class;
  TsHintManager = class;

  TsPassiveShadow = class(TPersistent)
  private
    FEnabled : boolean;
    FColor : TColor;
    FOffset : integer;
    FOwner : TsPassivePaintStyle;
    FTransparency : integer;
    FBlur : integer;

    procedure SetColor(const Value: TColor);
    procedure SetEnabled(const Value: boolean);
    procedure SetOffset(const Value: integer);
    procedure SetTransparency(const Value: integer);
    procedure SetBlur(const Value: integer);
  public
    {  }
    constructor Create(AOwner : TPersistent);
  published
    {   }
    property Transparency: integer read FTransparency write SetTransparency default 70;
    {      }
    property Enabled : boolean read FEnabled write SetEnabled default True;
    {  }
    property Color : TColor read FColor write SetColor default clBlack;
    {    }
    property Offset : integer read FOffset write SetOffset default 10;
    property Blur : integer read FBlur write SetBlur default 4;
  end;

  TsPassivePainting = class(TPersistent)
  private
    FColorActive : Graphics.TColor;
    FColorInActive : Graphics.TColor;
    FParentColor : boolean;
    FColorBorderTop : Graphics.TColor;
    FColorBorderBottom : Graphics.TColor;
    FTransparency: integer;
    FOwner : TsPassivePaintStyle;
    procedure SetColors (Index: Integer; Value: Graphics.TColor);
    procedure SetTransparency(const Value: integer);
  public
    {  }
    constructor Create(AOwner : TPersistent);
  published
    property Transparency: integer read FTransparency write SetTransparency default 0;
    {     }
    property ColorBorderTop: Graphics.TColor index 0 read FColorBorderTop write SetColors default scDarkSky;
    {     }
    property ColorBorderBottom: Graphics.TColor index 1 read FColorBorderBottom write SetColors default scDarkSky;
    {   ,   }
    property ColorActive: Graphics.TColor index 2 read FColorActive write SetColors default scLightYellow;
    {   ,    }
    property ColorInActive: Graphics.TColor index 3 read FColorInActive write SetColors default cl3DLight;
  end;

  TsPassivePaintStyle = class(TPersistent)
  protected
    FGroupIndex : integer;                             
  public
    FOwner : TComponent;
    FShadow : TsPassiveShadow;
    FPainting : TsPassivePainting;
    FDrawingStop : boolean;
    FCacheBmp : Graphics.TBitmap;
    COC : integer;
    sC : TsStyleControl;
    { , ,     }
    constructor Create(AOwner : TComponent); overload;
    destructor Destroy; override;
    procedure OwnerInvalidate;
    procedure WndProc(var Message: TMessage);
    procedure BeginDrawing;
    procedure EndDrawing;
  published
    {   @link(TsShadow)}
    property Shadow: TsPassiveShadow read FShadow write FShadow;
    { ,   }
    property Painting: TsPassivePainting read FPainting write FPainting;
    {  .   @link(TsStyleControl).  -1,     }
    property GroupIndex: integer read FGroupIndex write FGroupIndex default 0;
  end;

  TsHintManager = class(TComponent)
  private
    FsStyle : TsPassivePaintStyle;
    FBorderWidth : integer;
    FBlur: integer;
    FMarginH: integer;
    FMarginV : integer;
    FFont : TFont;
    FStyle : TsHintStyle;
    FRadius : integer;
    FExOffset : integer;
    FMaxWidth : integer;
    FPauseHide: integer;
    procedure SetFont(const Value: TFont);
    procedure SetStyle(const Value: TsHintStyle);
    procedure SetNewStyle(hs: TsHintStyle);
  protected
  public
    FForm: TCustomForm;
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Invalidate;
    procedure Assign(Source: TsHintManager); overload;
  published
    property Radius : integer read FRadius write FRadius default 10;
    property sStyle : TsPassivePaintStyle read FsStyle write FsStyle;
    property BorderWidth : integer read FBorderWidth write FBorderWidth default 1;
    property Blur : integer read FBlur write FBlur default 0;
    property ExOffset : integer read FExOffset write FExOffset default 8;
    property PauseHide : integer read FPauseHide write FPauseHide default 2500;
    property MarginH : integer read FMarginH write FMarginH default 3;
    property MarginV : integer read FMarginV write FMarginV default 3;
    property MaxWidth : integer read FMaxWidth write FMaxWidth default 200;
    property Style: TsHintStyle read FStyle write SetStyle default hsSimply;
    property Font: TFont read FFont write SetFont;
  end;

  TsHintWindow = class(THintWindow)
  private
  protected
    FMousePos : TsMousePosition;
    procedure Paint; override;
    procedure CreateParams(var Params: TCreateParams); override;
    function CalcHintRect(MaxWidth: Integer; const AHint: string; AData: Pointer): TRect; override;
    procedure WMEraseBkGND (var Message: TWMPaint); message WM_ERASEBKGND;
  public
    procedure ActivateHint(Rect: TRect; const AHint: string); override;
    constructor Create(AOwner:TComponent); override;
  end;

  function GetMousePosition : TPoint;

var
  Manager : TsHintManager;

implementation

uses sCustomMenuManager;

function GetMousePosition : TPoint;
var
  m : TMouse;
begin
  m := TMouse.Create;
  Result := m.CursorPos;
  inc(Result.x, 0);
  inc(Result.y, 0);
  m.Free;
end;

{ TsPassiveShadow }

constructor TsPassiveShadow.Create(AOwner: TPersistent);
begin
  inherited Create;
  FOwner := TsPassivePaintStyle(AOwner);
  FColor := clBlack;
  FBlur := 4;
  FOffset := 10;
  FTransparency := 70;
  FEnabled := True;
end;

procedure TsPassiveShadow.SetBlur(const Value: integer);
begin
  if FBlur <> Value then begin
    FBlur := Value;
  end;
end;

procedure TsPassiveShadow.SetColor(const Value: TColor);
begin
  if (FColor <> Value) then begin
    FColor := Value;
  end;
end;

procedure TsPassiveShadow.SetEnabled(const Value: boolean);
begin
  if (FEnabled <> Value) then begin
    FEnabled := Value;
  end;
end;

procedure TsPassiveShadow.SetOffset(const Value: integer);
begin
  if (FOffset <> Value) then begin
    FOffset := Value;
  end;
end;

procedure TsPassiveShadow.SetTransparency(const Value: integer);
begin
  if FTransparency <> Value then begin
    if Value < 0 then FTransparency := 0
    else if Value > 100 then FTransparency := 100
    else FTransparency := Value;
  end;
end;

{ TsPassivePainting }

constructor TsPassivePainting.Create(AOwner: TPersistent);
begin
  inherited Create;
  FColorBorderTop := scDarkSky;
  FColorBorderBottom := scDarkSky;
  FColorActive := scLightYellow;
  FColorInActive := scLightBlue;
  FOwner := TsPassivePaintStyle(AOwner);
end;

procedure TsPassivePainting.SetColors(Index: Integer; Value: TColor);
begin
  case Index of
    0: FColorBorderTop := Value;
    1: FColorBorderBottom := Value;
    2: FColorActive := Value;
    3: begin
      FColorInActive := Value;
    end;
  end;
end;

procedure TsPassivePainting.SetTransparency(const Value: integer);
begin
  if FTransparency <> Value then begin
    FTransparency := Value;
  end;
end;

{ TsPassivePaintStyle }

constructor TsPassivePaintStyle.Create(AOwner : TComponent);
begin
  inherited Create;
  FOwner := AOwner;
  FCacheBmp := Graphics.TBitmap.Create;

  Shadow := TsPassiveShadow.Create(Self);
  Painting := TsPassivePainting.Create(Self);

{$IFDEF ALLOWPROP}
  sC := GetsStyleControl(sVclUtils.GetParentForm(AOwner), GroupIndex);
  if not Assigned(sC) then begin
    FGroupIndex := 0;
  end
  else begin
    Painting.FColorActive := sC.ColorActive;
    Painting.FColorInActive := sC.ColorInActive;
    Painting.FColorBorderTop := sC.ColorBorderTop;
    Painting.FColorBorderBottom := sC.ColorBorderBottom;
    FGroupIndex := sC.GroupIndex;
    Painting.FParentColor := sC.ParentColor;
    Shadow.Color := sC.ShadowColor;
    Shadow.Offset := sC.ShadowOffset;
    Shadow.Transparency := sC.ShadowTransparency;
    Shadow.Blur := sC.ShadowBlur;
    Shadow.Enabled := sC.ShadowEnabled;
  end;
{$ENDIF}
end;

destructor TsPassivePaintStyle.Destroy;
begin
  Painting.Free;
  Shadow.Free;
  inherited;
end;                   

procedure TsPassivePaintStyle.OwnerInvalidate;
begin
  Case COC of
    COC_TsHintManager : begin
      TsHintManager(FOwner).Invalidate;
    end;
    COC_TsCustomMenuManager, COC_TsMainMenuManager : begin
      TsCustomMenuManager(FOwner).Invalidate;
    end;
  end;
end;

procedure TsPassivePaintStyle.BeginDrawing;
begin
  FDrawingStop := True;
end;

procedure TsPassivePaintStyle.EndDrawing;
begin
  FDrawingStop := False;
end;

procedure TsPassivePaintStyle.WndProc(var Message: TMessage);
var
  i : integer;
  sSC : TsStyleControl;
begin
{$IFDEF ALLOWPROP}
  if (Message.Msg >= SM_OFFSET) and (Message.Msg <= SM_LAST) then begin
    if ((Message.WParam = GroupIndex) or (Message.WParam = GI_FORPANELPATTERN)) then begin
      case Message.Msg of
        SM_SETBORDERTOP:      Painting.ColorBorderTop := TSMSetColor(Message).Value;
        SM_SETBORDERBOTTOM:   Painting.ColorBorderBottom := TSMSetColor(Message).Value;
        SM_SETCOLORACTIVE:    Painting.ColorActive := TSMSetColor(Message).Value;
        SM_SETCOLORINACTIVE:  Painting.ColorInActive := TSMSetColor(Message).Value;
        SM_SETCOLORSHADOW :   Shadow.Color := TSMSetColor(Message).Value;
        SM_SETPATTERN: begin  end;

        SM_SETSHADOWENABLED:  Shadow.Enabled := TSMSetBoolean(Message).Value;
        SM_SHADOWTRANSPARENCY:Shadow.Transparency := TSMSetInteger(Message).Value;
        SM_SHADOWOFFSET:      Shadow.Offset       := TSMSetInteger(Message).Value * 2;
        SM_SHADOWBLUR:        Shadow.Blur         := TSMSetInteger(Message).Value * 2;

        SM_GETSTYLEINFO:      begin
            TSMGetStyleInfo(Message).WParam := 1;
            TSMGetStyleInfo(Message).LParam := Longint(Self);
          end;

        SM_CHANGESTYLECONTROL : begin
            BeginDrawing;
            sSC := TSMChangeStyleControl(Message).sStyleControl;

            Painting.ColorBorderTop := sSC.ColorBorderTop;
            Painting.ColorBorderBottom := sSC.ColorBorderBottom;
            Painting.ColorActive := sSC.ColorActive;
            Painting.ColorInActive := sSC.ColorInActive;
            Shadow.Color := sSC.ShadowColor;
            Shadow.Enabled := sSC.ShadowEnabled;

            Shadow.Offset       := sSC.ShadowOffset * 2;
            Shadow.Blur         := sSC.ShadowBlur * 2;
            EndDrawing;
          end;

      end;
    end;
  end;
{$ENDIF}  
end;

{ TsHintManager }

procedure TsHintManager.Assign(Source: TsHintManager);
begin
  FFont.Assign(Source.Font);
  FBorderWidth := Source.BorderWidth;
  FBlur := Source.Blur;
  FMarginH := Source.MarginH;
  FMarginV := Source.MarginV;
  FRadius := Source.Radius;
  sStyle.Painting.Transparency := Source.sStyle.Painting.Transparency;
  sStyle.Painting.ColorActive := Source.sStyle.Painting.ColorActive;
  sStyle.Painting.ColorBorderTop := Source.sStyle.Painting.ColorBorderTop;
  sStyle.Painting.ColorBorderBottom := Source.sStyle.Painting.ColorBorderBottom;
  sStyle.Shadow.FEnabled := Source.sStyle.Shadow.Enabled;
  sStyle.Shadow.FOffset := Source.sStyle.Shadow.Offset;
  sStyle.Shadow.FTransparency := Source.sStyle.Shadow.Transparency;
end;

constructor TsHintManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FsStyle := TsPassivePaintStyle.Create(Self);
  FForm := sVclUtils.GetParentForm(AOwner);
  FFont := TFont.Create;
  FFont.Size := 10;
  FFont.Style := [fsBold];
  FFont.Color := scDarkSky;
  sStyle.COC := COC_TsHintManager;
  FBorderWidth := 1;
  FBlur := 0;
  FMarginH := 10;
  FMarginV := 5;
  FRadius := 10;
  FStyle := hsSimply;
  FExOffset := 8;
  FMaxWidth := 200;
  FPauseHide := 2500;

  Application.HintPause := 500;
  Application.HintShortPause := 200;
  invalidate;
end;

destructor TsHintManager.Destroy;
begin
  FsStyle.Free;
  FFont.Free;
  inherited;
end;

procedure TsHintManager.Invalidate;
begin
  if Assigned(sStyle) then begin
    SetNewStyle(Style);
  end;
end;

procedure TsHintManager.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TsHintManager.SetNewStyle(hs: TsHintStyle);
begin
  if not (csDesigning in ComponentState) then begin
    if hs = hsStandard then begin
      HintWindowClass := THintWindow;
    end
    else begin
      Manager := Self;
      HintWindowClass := TsHintWindow;
    end;
  end;
end;

procedure TsHintManager.SetStyle(const Value: TsHintStyle);
begin
  FStyle := Value;
  SetNewStyle(Value);
end;

{ TsHintWindow }

procedure TsHintWindow.ActivateHint(Rect: TRect; const AHint: string);
var
  DC: HDC;
  w, h : integer;
  t, l : boolean;
  p : TPoint;
begin
  if (Manager.Style = hsNone) or not Assigned(Manager) then exit;

  Application.HintHidePause := Manager.FPauseHide;
  Application.ProcessMessages;
  ShowWindow(Handle, SW_HIDE);
  Caption := AHint;
  p := GetMousePosition;

  w := Rect.Right - Rect.Left;
  h := Rect.Bottom - Rect.Top;
  FMousePos := mpLeftTop;

{
  if Manager.Style  = hsExtended then begin
    Rect.Left := p.x;
    Rect.Top := p.y;
    Rect.Right := p.x + w;
    Rect.Bottom := p.y + h;
  end;
}
  UpdateBoundsRect(Rect);

  case Manager.Style of
    hsSimply : begin
      if Rect.Bottom > Screen.DesktopHeight then begin
        Rect.Top := Screen.DesktopHeight - h - 2;
      end;
      if Rect.Top < Screen.DesktopTop then begin
        Rect.Top := Screen.DesktopTop + 2;
      end;
      if Rect.Right > Screen.DesktopWidth then begin
        Rect.Left := Screen.DesktopWidth - w - 2;
      end;
      if Rect.Left < Screen.DesktopLeft then begin
        Rect.Left := Screen.DesktopLeft + 2;
      end;
    end;
{    hsExtended : begin
      t := True;
      l := True;
      if Rect.Bottom > Screen.DesktopHeight then begin
        Rect.Top := p.y - h;
        t := False;
      end;
      if Rect.Top < Screen.DesktopTop then begin
        Rect.Top := p.y;
        t := True;
      end;
      if Rect.Right > Screen.DesktopWidth then begin
        Rect.Left := p.x - w;
        l := False;
      end;
      if Rect.Left < Screen.DesktopLeft then begin
        Rect.Left := p.x;
        l := True;
      end;
      if t and l         then FMousePos := mpLeftTop;
      if t and not l     then FMousePos := mpRightTop;
      if not t and l     then FMousePos := mpLeftBottom;
      if not t and not l then FMousePos := mpRightBottom;
    end;
}
  end;

  Rect.Right := Rect.Left + w;
  Rect.Bottom := Rect.Top + h;

  Manager.sStyle.FCacheBmp.Width := w;
  Manager.sStyle.FCacheBmp.Height := h;

  DC := GetDC(0);

  // Grabbing of screen
  BitBlt(Manager.sStyle.FCacheBmp.Canvas.Handle,
               0,
               0,
               w,
               h,
               DC,
               Rect.Left,
               Rect.Top,
               SrcCopy);

  ReleaseDC(0, DC);

  // Show window with hint
  SetWindowPos(Handle,
               HWND_TOPMOST,
               Rect.Left,
               Rect.Top,
               w,
               h,
               SWP_SHOWWINDOW or SWP_NOACTIVATE);
  Invalidate;
end;

constructor TsHintWindow.Create(AOwner: TComponent);
begin
  inherited;
end;

procedure TsHintWindow.CreateParams(var Params: TCreateParams);
begin
  inherited;
  Params.Style := Params.Style - WS_BORDER + WS_EX_TRANSPARENT;
end;

procedure TsHintWindow.Paint;
var
  P : TPoint;
  R, rText: TRect;
  SmallOffset, ShadowOffset : integer;
  function GetTextRect : TRect; begin
    Result := ClientRect;
    case FMousePos of
      mpLeftTop: begin
        inc(Result.Top, Manager.FExOffset + Manager.BorderWidth);
        inc(Result.Left, Manager.FExOffset + Manager.BorderWidth);
        dec(Result.Right, ShadowOffset + Manager.BorderWidth);
        dec(Result.Bottom, ShadowOffset + Manager.BorderWidth);
      end;
      mpLeftBottom: begin
        dec(Result.Bottom, Manager.FExOffset + Manager.BorderWidth);
        inc(Result.Left, Manager.FExOffset + Manager.BorderWidth);
        dec(Result.Right, ShadowOffset + Manager.BorderWidth);
        inc(Result.Top, ShadowOffset + Manager.BorderWidth);
      end;
      mpRightTop: begin
        inc(Result.Top, Manager.FExOffset + Manager.BorderWidth);
        dec(Result.Right, Manager.FExOffset + Manager.BorderWidth);
        inc(Result.Left, ShadowOffset + Manager.BorderWidth);
        dec(Result.Bottom, ShadowOffset + Manager.BorderWidth);
      end;
      mpRightBottom: begin
        dec(Result.Bottom, Manager.FExOffset + Manager.BorderWidth);
        dec(Result.Right, Manager.FExOffset + Manager.BorderWidth);
        inc(Result.Left, ShadowOffset + Manager.BorderWidth);
        inc(Result.Top, ShadowOffset + Manager.BorderWidth);
      end;
    end;
  end;
begin
  R := ClientRect;
  ShadowOffset := iffi(Manager.sStyle.Shadow.Enabled, Manager.sStyle.Shadow.Offset, 0);

  case Manager.Style of
    hsSimply : begin
      // Shadow drawing
      if Manager.sStyle.Shadow.Enabled then begin
        FadeBmp(Manager.sStyle.FCacheBmp,
                Classes.Rect(R.Left + ShadowOffset,
                R.Top + ShadowOffset, R.Right, R.Bottom),
                Manager.sStyle.Shadow.Transparency, Manager.sStyle.Shadow.Color, Manager.sStyle.Shadow.Blur, 0);
      end;

      case Manager.sStyle.Painting.Transparency of
        0: begin
          Manager.sStyle.FCacheBmp.Canvas.Brush.Style := bsSolid;
          Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psClear;
          Manager.sStyle.FCacheBmp.Canvas.Brush.Color := Manager.sStyle.Painting.ColorActive;
          Manager.sStyle.FCacheBmp.Canvas.Rectangle(
                  Classes.Rect(R.Left, R.Top, R.Right - ShadowOffset, R.Bottom - ShadowOffset)
                 );
        end;
        else begin
          FadeBmp(Manager.sStyle.FCacheBmp,
                  Classes.Rect(R.Left, R.Top, R.Right - ShadowOffset, R.Bottom - ShadowOffset),
                  Manager.sStyle.Painting.Transparency, Manager.sStyle.Painting.ColorActive, Manager.Blur, Manager.Radius);
        end;
      end;

      // Text
//      InflateRect(R, - Manager.MarginV - Manager.FBorderWidth , - Manager.MarginH - Manager.FBorderWidth);
      inc(R.Top, Manager.MarginV + Manager.BorderWidth);
//      inc(R.Left, Manager.MarginH);
//      dec(R.Right, Manager.MarginH);

      Manager.sStyle.FCacheBmp.Canvas.Brush.Style := bsClear;
      DrawText(Manager.sStyle.FCacheBmp.Canvas.Handle, PChar(Caption), -1, R, DT_CENTER or DT_NOPREFIX or
        DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly);

      R := Classes.Rect(ClientRect.Left,
                        ClientRect.Top,
                        ClientRect.Right - ShadowOffset,
                        ClientRect.Bottom - ShadowOffset);

      if Manager.FBorderWidth > 0 then begin
        Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psSolid;
        Frame3d(Manager.sStyle.FCacheBmp.Canvas,
                R,
                Manager.sStyle.Painting.ColorBorderTop, Manager.sStyle.Painting.ColorBorderBottom, Manager.FBorderWidth);
      end;

      // Copy canvas
      BitBlt(Canvas.Handle,
             0,
             0,
             Manager.sStyle.FCacheBmp.Width,
             Manager.sStyle.FCacheBmp.Height,
             Manager.sStyle.FCacheBmp.Canvas.Handle,
             0,
             0,
             SRCCOPY);
    end;
{
    hsExtended : begin

      rText := GetTextRect;

      case Manager.sStyle.Painting.Transparency of
        0: begin
          Manager.sStyle.FCacheBmp.Canvas.Brush.Style := bsSolid;
          Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psSolid;
          Manager.sStyle.FCacheBmp.Canvas.Pen.Width := Manager.BorderWidth;
          Manager.sStyle.FCacheBmp.Canvas.Brush.Color := Manager.sStyle.Painting.ColorActive;
          Manager.sStyle.FCacheBmp.Canvas.Pen.Color := Manager.sStyle.Painting.ColorBorderBottom;

          Manager.sStyle.FCacheBmp.Canvas.RoundRect(
                  rText.Left, rText.Top,
                  rText.Right, rText.Bottom,
                  Manager.Radius, Manager.Radius
                 );

          Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psClear;
//          inc(SmallOffset, Manager.Radius);
          SmallOffset := Manager.FExOffset div 10 + 6;
          case FMousePos of
            mpLeftTop : begin
              Manager.sStyle.FCacheBmp.Canvas.Polygon([
                Point(0, 0),
                Point(rText.Left + SmallOffset, rText.Top),
                Point(rText.Left + SmallOffset, rText.Top + SmallOffset),
                Point(rText.Left, rText.Left + SmallOffset)
              ]);
              Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psSolid;
              Manager.sStyle.FCacheBmp.Canvas.PolyLine([
                Point(rText.Left, rText.Top + SmallOffset),
                Point(0, 0),
                Point(rText.Left + SmallOffset, rText.Top)
              ]);
            end;
            mpLeftBottom : begin
              Manager.sStyle.FCacheBmp.Canvas.Polygon([
                Point(0, Height),
                Point(rText.Left + SmallOffset, rText.Bottom),
                Point(rText.Left + SmallOffset, rText.Bottom - SmallOffset),
                Point(rText.Left, rText.Bottom - SmallOffset)
              ]);
              Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psSolid;
              Manager.sStyle.FCacheBmp.Canvas.PolyLine([
                Point(rText.Left + SmallOffset, rText.Bottom),
                Point(0, Height),
                Point(rText.Left, rText.Bottom - SmallOffset)
              ]);
            end;
            mpRightBottom : begin
              Manager.sStyle.FCacheBmp.Canvas.Polygon([
                Point(Width, Height),
                Point(rText.Right - SmallOffset, rText.Bottom),
                Point(rText.Right - SmallOffset, rText.Bottom - SmallOffset),
                Point(rText.Right, rText.Bottom - SmallOffset)
              ]);
              Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psSolid;
              Manager.sStyle.FCacheBmp.Canvas.PolyLine([
                Point(rText.Right - SmallOffset, rText.Bottom),
                Point(Width, Height),
                Point(rText.Right, rText.Bottom - SmallOffset)
              ]);
            end;
            mpRightTop : begin
              Manager.sStyle.FCacheBmp.Canvas.Polygon([
                Point(Width, 0),
                Point(rText.Right - SmallOffset, rText.Top),
                Point(rText.Right - SmallOffset, rText.Top + SmallOffset),
                Point(rText.Right, rText.Top + SmallOffset)
              ]);
              Manager.sStyle.FCacheBmp.Canvas.Pen.Style := psSolid;
              Manager.sStyle.FCacheBmp.Canvas.PolyLine([
                Point(rText.Right - SmallOffset, rText.Top),
                Point(Width, 0),
                Point(rText.Right, rText.Top + SmallOffset)
              ]);
            end;
          end;
        end;
      end;

      // Text
      inc(rText.Top, Manager.FMarginV);
      inc(rText.Left, Manager.FMarginH);
      dec(rText.Bottom, Manager.FMarginV);
      dec(rText.Right, Manager.FMarginH);
      Manager.sStyle.FCacheBmp.Canvas.Brush.Style := bsClear;
      DrawText(Manager.sStyle.FCacheBmp.Canvas.Handle, PChar(Caption), -1,
                  rText, DT_CENTER or DT_NOPREFIX or DT_VCENTER or
                         DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly
                         );

      // Copy canvas
      BitBlt(Canvas.Handle,
             0,
             0,
             Manager.sStyle.FCacheBmp.Width,
             Manager.sStyle.FCacheBmp.Height,
             Manager.sStyle.FCacheBmp.Canvas.Handle,
             0,
             0,
             SRCCOPY);
    end;
}    
  end;
end;

function TsHintWindow.CalcHintRect(MaxWidth: Integer; const AHint: string; AData: Pointer): TRect;
begin
  Result := Rect(0, 0, Manager.FMaxWidth, 0);
  if Assigned(Manager) and Assigned(Manager.Font) then Manager.sStyle.FCacheBmp.Canvas.Font.Assign(Manager.Font);

  DrawText(Manager.sStyle.FCacheBmp.Canvas.Handle, PChar(AHint), -1, Result,
           DT_CALCRECT or DT_CENTER or DT_VCENTER or DT_WORDBREAK or DT_NOPREFIX or DrawTextBiDiModeFlagsReadingOnly);

  Inc(Result.Right, (Manager.MarginH + Manager.FBorderWidth) * 2 + iffi(Manager.sStyle.Shadow.Enabled, Manager.sStyle.Shadow.Offset, 0));
  Inc(Result.Bottom, (Manager.MarginV + Manager.FBorderWidth) * 2 + iffi(Manager.sStyle.Shadow.Enabled, Manager.sStyle.Shadow.Offset, 0));

//  Inc(Result.Right, integer(Manager.Style = hsExtended) * Manager.FExOffset);
//  Inc(Result.Bottom, integer(Manager.Style = hsExtended) * Manager.FExOffset);
end;

procedure TsHintWindow.WMEraseBkGND(var Message: TWMPaint);
begin
  Message.Result := 1;
end;

end.
